;; Generated package description from vc-backup.el  -*- no-byte-compile: t -*-
(define-package "vc-backup" "1.1.0.0.20211102.100756" "VC backend for versioned backups" 'nil :authors '(("Philip Kaludercic" . "philipk@posteo.net")) :maintainer '("Philip Kaludercic" . "philipk@posteo.net") :keywords '("vc") :url "https://git.sr.ht/~pkal/vc-backup")
